from protein_structure import ProteinStructure
import protein_structure_utils
from pathlib import Path

# Test the ProteinStructure class
protein = ProteinStructure()
protein.read_pdb("example.pdb", name="test")

# Print the protein structure
for chain in protein.chains:
    print(f"Chain {chain.name}")
    for residue in chain.residues:
        print(f"Residue {residue.name} {residue.res_seq}")
        for atom in residue.atoms:
            print(f"Atom {atom.name} at {atom.x}, {atom.y}, {atom.z}")

protein = ProteinStructure()
protein.read_pdb("example.pdb")

# Display the amino acid sequence of the protein
# Get the sequence of each chain in the protein
sequence_dict = protein.get_sequence_dict()
for chain_id, sequence in sequence_dict.items():
    print(f"Chain {chain_id}: {sequence}")

# What is the radius of gyration of the backbone of chain A in our protein structure?
rgA = protein_structure_utils.get_radgyr(protein, chains='A', atom_type="backbone")
print(f"protein structure chain A rg : {rgA:.4f}")

# What is the radius of gyration of the backbone of an ideal alanine helix?
ideal_helix_seq_length = len(sequence_dict['A'])
rg_ideal_helix = protein_structure_utils.get_radgyr_alanine_helix(ideal_helix_seq_length, atom_type="backbone")
print(f"ideal alanine helix rg : {rg_ideal_helix:.4f}")

# What is the radius of gyration ratio of the protein structure to an ideal alanine helix?
rg_ratio = protein_structure_utils.get_radgyr_ratio(protein, chains='A', atom_type="backbone")
print(f"rg ratio : {rg_ratio:.4f}")


import unittest

import sys
sys.path.append('../protein_design_tools')
from protein_structure import ProteinStructure
from protein_structure_utils import get_coordinates

class TestProteinStructureUtils(unittest.TestCase):
    def setUp(self):
        self.protein = ProteinStructure()
        self.protein.read_pdb("example.pdb", name="test")

    def test_get_coordinates_all(self):
        all_coordinates = get_coordinates(self.protein, atom_type="all")
        # Add assertions here to check the result
        # For example:
        self.assertIsNotNone(all_coordinates)
        # Add more assertions based on your expected results

    def test_get_coordinates_backbone(self):
        backbone_coordinates = get_coordinates(self.protein, atom_type="backbone")
        # Add assertions here to check the result
        # For example:
        self.assertIsNotNone(backbone_coordinates)
        # Add more assertions based on your expected results

    def test_get_coordinates_sidechain(self):
        sidechain_coordinates = get_coordinates(self.protein, atom_type="sidechain")
        # Add assertions here to check the result
        # For example:
        self.assertIsNotNone(sidechain_coordinates)
        # Add more assertions based on your expected results

if __name__ == '__main__':
    unittest.main()



"""
protein_structure_utils.py
====================================
The protein_structure_utils module contains functions for analyzing protein structures.
"""
import numpy as np
from pathlib import Path

# make a flexible function that returns coordinates from a structure. Function should be able to read all atom coordinates, or cordinates of backbone atoms, or cordinates of CA atoms, or coordinates from specified chains, or from specified residue numbers
def get_coordinates(structure, atom_type="all", chains=None, residue_numbers=None, residue_indices=None, residue_ids=None):
    """
    Get the coordinates of atoms from a ProteinStructure object.
    
    Parameters
    ----------
    structure : ProteinStructure
        The protein structure to get the coordinates from.
    atom_type : str, optional
        The type of atoms to get the coordinates from. Can be "all", "backbone", or "sidechain". Defaults to "all".
    chains : str or list of str, optional
        A list of chain names to get the coordinates from, or a single chain as a string. If None, get the coordinates 
        from all chains. Defaults to None.
    residue_numbers : list of int, optional
        A list of residue numbers to get the coordinates from. If None, get the coordinates from all residues. Defaults 
        to None.
    residue_indices : list of int, optional
        A list of residue sequence indices to get the coordinates from. If None, get the coordinates from all residues. 
        Defaults to None.
    """
    # If the user passed a string for chains, convert it to a list
    if isinstance(chains, str):
        chains = [chains]

    coordinates = [] # Initialize an empty list to store the coordinates
    
    # Loop over all chains in the structure
    for chain in structure.chains:
        # Check if the chain is in the specified chains
        if chains is not None and chain.name not in chains:
            continue
        
        # Loop over all residues in the chain
        for residue in chain.residues:
            # Check if the residue number is in the specified residue numbers
            if residue_numbers is not None and residue.res_seq not in residue_numbers:
                continue
            
            # Loop over all atoms in the residue append the atom to the coordinates list depending on the atom type
            for atom in residue.atoms:
                if atom_type == "all" or atom.name == atom_type:
                    coordinates.append([atom.x, atom.y, atom.z])
                elif atom_type == "backbone" and atom.name in ["N", "CA", "C", "O"]:
                    coordinates.append([atom.x, atom.y, atom.z])
                elif atom_type == "sidechain" and atom.name not in ["N", "CA", "C", "O"]:
                    coordinates.append([atom.x, atom.y, atom.z])

    coordinates = np.array(coordinates) # Convert the list of coordinates to a numpy array
    
    return coordinates

def get_masses(structure, atom_type="all", chains=None, residue_numbers=None, residue_indices=None, residue_ids=None):
    """
    Get the masses of atoms from a ProteinStructure object.
    
    Parameters
    ----------
    structure : ProteinStructure
        The protein structure to get the masses from.
    atom_type : str, optional
        The type of atoms to get the masses from. Can be "all", "backbone", or "sidechain". Defaults to "all".
    chains : str or list of str, optional
        A list of chain names to get the masses from, or a single chain as a string. If None, get the masses from all 
        chains. Defaults to None.
    residue_numbers : list of int, optional
        A list of residue numbers to get the masses from. If None, get the masses from all residues. Defaults to None.
    residue_indices : list of int, optional
        A list of residue sequence indices to get the masses from. If None, get the masses from all residues. Defaults 
        to None.
    residue_ids : list of str, optional
        A list of residue sequence identifiers to get the masses from. If None, get the masses from all residues. 
        Defaults to None.
    """
    # If the user passed a string for chains, convert it to a list
    if isinstance(chains, str):
        chains = [chains]

    masses = [] # Initialize an empty list to store the masses
    
    # Loop over all chains in the structure
    for chain in structure.chains:
        # Check if the chain is in the specified chains
        if chains is not None and chain.name not in chains:
            continue
        
        # Loop over all residues in the chain
        for residue in chain.residues:
            # Check if the residue number is in the specified residue numbers
            if residue_numbers is not None and residue.res_seq not in residue_numbers:
                continue
            
            # Loop over all atoms in the residue append the atom to the masses list depending on the atom type
            for atom in residue.atoms:
                if atom_type == "all" or atom.name == atom_type:
                    masses.append(atom.mass)
                elif atom_type == "backbone" and atom.name in ["N", "CA", "C", "O"]:
                    masses.append(atom.mass)
                elif atom_type == "sidechain" and atom.name not in ["N", "CA", "C", "O"]:
                    masses.append(atom.mass)

    return masses

def get_radgyr(structure, atom_type="backbone", chains=None, residue_numbers=None, residue_indices=None, residue_ids=None):
    """
    Calculate the radius of gyration of a protein structure.

    Parameters:
    structure (ProteinStructure): The protein structure to calculate the radius of gyration for.
    atom_type (str, optional): The type of atoms to calculate the radius of gyration for. Can be "all", "backbone", or "sidechain". Defaults to "backbone".
    chains (list of str, optional): A list of chain names to calculate the radius of gyration for. If None, calculate the radius of gyration for all chains. Defaults to None.
    residue_numbers (list of int, optional): A list of residue numbers to calculate the radius of gyration for. If None, calculate the radius of gyration for all residues. Defaults to None.
    residue_indices (list of int, optional): A list of residue sequence indices to calculate the radius of gyration for. If None, calculate the radius of gyration for all residues. Defaults to None.
    residue_ids (list of str, optional): A list of residue sequence identifiers to calculate the radius of gyration for. If None, calculate the radius of gyration for all residues. Defaults to None.
    
    Returns:
    float: The radius of gyration of the selected atoms.
    """

    # If the user passed a string for chains, convert it to a list
    if isinstance(chains, str):
        chains = [chains]

    # Get the coordinates of all atoms in the structure using the 


    coordinates = get_coordinates(structure, atom_type=atom_type, chains=chains, residue_numbers=residue_numbers, residue_indices=residue_indices, residue_ids=residue_ids)
    masses = get_masses(structure, atom_type=atom_type, chains=chains, residue_numbers=residue_numbers, residue_indices=residue_indices, residue_ids=residue_ids)
    
    # Calculate the center of mass
    center_of_mass = np.average(coordinates, axis=0, weights=masses)

    # Calculate the distance of each atom from the center of mass
    distances = np.linalg.norm(coordinates - center_of_mass, axis=1)

    # Calculate the radius of gyration
    radius_of_gyration = np.sqrt(np.average(distances**2, weights=masses))

    return radius_of_gyration

# get radius of gyration for all alanine helices of different lengths
# structure, atom_type="all", chains=None, residue_numbers=None, residue_indices=None, residue_ids=None):
def get_radgyr_alanine_helix(sequence_length, atom_type="backbone"):
    """
    Calculate the radius of gyration of an ideal alanine helix. The radius of gyration is calculated from a linear 
    regression model performed on a dataset of ideal alanine helices of different lengths.
    
    Parameters
    ----------
    sequence_length : int
        The length of the alanine helix.
    atom_type : str, optional
        The type of atoms to calculate the radius of gyration for. Can be "all" or "backbone". Defaults to "backbone".
    """
    if atom_type == "all":
        # values are from a linear regression of a dataset of ideal alanine helices of different lengths
        radius_of_gyration = 0.4704427974768968 * float(sequence_length) + 0.33788445679512336
    elif atom_type == "backbone":
        # values are from a linear regression of a dataset of ideal alanine helices of different lengths
        radius_of_gyration = 0.4706800060715209 * float(sequence_length) + 0.17986881616526773
    else:
        raise ValueError("Invalid atom type. Atom type must be 'all' or 'backbone'.")
    
    return radius_of_gyration

def get_radgyr_ratio(structure, atom_type="backbone", chains=None):
    """
    Calculate the radius of gyration ratio of a protein structure to an ideal alanine helix of the same length.
    
    Parameters
    ----------
    structure : ProteinStructure
        The protein structure to calculate the radius of gyration ratio for.
    atom_type : str, optional
        The type of atoms to calculate the radius of gyration for. Can be "all" or "backbone". Defaults to "backbone".
    chains : list of str, optional
        A list of chain names to calculate the radius of gyration for. If None, calculate the radius of gyration for 
        all chains. Defaults to None.
    """

    # If the user passed a string for chains, convert it to a list
    if isinstance(chains, str):
        chains = [chains]

    # Get the radius of gyration
    structure_radius_of_gyration = get_radgyr(structure, atom_type=atom_type, chains=chains)

    # get the sequence length of the protein
    alanine_helix_length = 0
    sequence_dict = structure.get_sequence_dict()
    for chain in sequence_dict.keys():
        if chains is None or chain in chains:
            alanine_helix_length += len(sequence_dict[chain])
        else:
            continue

    alanine_radius_of_gyration = get_radgyr_alanine_helix(alanine_helix_length, atom_type=atom_type)

    # Calculate the radius of gyration ratio
    radius_of_gyration_ratio = structure_radius_of_gyration / alanine_radius_of_gyration

    return radius_of_gyration_ratio



"""
protein_structure.py
====================================
The protein_structure module contains the ProteinStructure class, which represents a protein structure and its 
components.
"""
from pathlib import Path
import h5py
import numpy as np

# Atomic weights from https://physics.nist.gov/cgi-bin/Compositions/stand_alone.pl?ele=&ascii=ascii
ATOMIC_WEIGHTS = {
    'H': 1.00794075405578,
    'C': 12.0107358967352,
    'N': 14.0067032114458,
    'O': 15.9994049243183,
    'S': 32.0647874061271
    # Add the rest of the elements here
}

# Dictionary to convert three-letter amino acid codes to one-letter codes
THREE_TO_ONE = {'ALA': 'A',
                'ARG': 'R',
                'ASN': 'N',
                'ASP': 'D',
                'CYS': 'C',
                'GLN': 'Q',
                'GLU': 'E',
                'GLY': 'G',
                'HIS': 'H',
                'ILE': 'I',
                'LEU': 'L',
                'LYS': 'K',
                'MET': 'M',
                'PHE': 'F',
                'PRO': 'P',
                'SER': 'S',
                'THR': 'T',
                'TRP': 'W',
                'TYR': 'Y',
                'VAL': 'V'
                }

ATOM_NAME_TO_ELEMENT = {'N': 'N',
                        'CA': 'C',
                        'C': 'C',
                        'O': 'O',
                        'CB': 'C',
                        'CG': 'C',
                        'CG1': 'C',
                        'CG2': 'C',
                        'CD': 'C',
                        'CD1': 'C',
                        'CD2': 'C',
                        'CE': 'C',
                        'CE1': 'C',
                        'CE2': 'C',
                        'CE3': 'C',
                        'CZ': 'C',
                        'CZ2': 'C',
                        'CZ3': 'C',
                        'CH2': 'C',
                        'ND1': 'N',
                        'ND2': 'N',
                        'NE': 'N',
                        'NE1': 'N',
                        'NE2': 'N',
                        'NH1': 'N',
                        'NH2': 'N',
                        'NZ': 'N',
                        'SD': 'S',
                        'SG': 'S',
                        'OG': 'O',
                        'OG1': 'O',
                        'OD1': 'O',
                        'OD2': 'O',
                        'OE1': 'O',
                        'OE2': 'O',
                        'OH': 'O',
                        'OXT': 'O',
                        'H': 'H'}

class ProteinStructure:
    """ProteinStruture represents a protein structure and its components."""
    def __init__(self, name=None):
        """
        Initialize a ProteinStructure object.

        Parameters
        ----------
        name : str, optional
            The name of the protein structure. Defaults to None.
        """
        self.name = None
        self.chains = []

    class Chain:
        """Chain represents a chain in a protein structure."""
        def __init__(self, name):
            """
            Initialize a Chain object, which will contain a list of Residue objects.
            
            Parameters
            ----------
            name : str
                chainID - chain identifier
            """
            self.name = name
            self.residues = [] # list of Residue objects

        class Residue:
            """Represents a residue in a protein structure."""
            # Residue class will contain a list of Atom objects
            def __init__(self, name, res_seq, i_code):
                """
                Initialize a Residue object, which will contain a list of Atom objects.
                
                Parameters
                ----------
                res_name : str
                    The name of the residue.
                res_seq : int
                    The sequence number of the residue.
                i_code : str
                    The insertion code of the residue.
                """
                self.name = name
                self.res_seq = res_seq
                self.i_code = i_code

                # Unique identifiers for each residue
                self.index = None
                self.res_seq_i = f"{res_seq}{i_code}"
                self.res_name_seq_i = f"{name}{res_seq}{i_code}"

                self.atoms = [] # list of Atom objects

            class Atom:
                """Represents an atom in a protein structure."""
                def __init__(self, atom_id, name, alt_loc, x, y, z, occupancy, temp_factor, segment_id, element, charge):
                    """
                    Initialize an Atom object.
                    
                    Parameters
                    ----------
                    atom_id : int
                        The atom serial number.
                    name : str
                        The atom  name.
                    alt_loc : str
                        The alternate location indicator.
                    x : float
                        The orthogonal coordinates for X in Angstroms.
                    y : float
                        The orthogonal coordinates for Y in Angstroms.
                    z : float
                        The orthogonal coordinates for Z in Angstroms.
                    occupancy : float
                        The occupancy of the atom.
                    temp_factor : float
                        The temperature factor of the atom.
                    segment_id : str
                        The segment ID.
                    element : str
                        The element symbol of the atom.
                    charge : str
                        The charge on the atom.
                    """
                    self.atom_id = atom_id
                    self.name = name
                    self.alt_loc = alt_loc
                    self.x = x
                    self.y = y
                    self.z = z
                    self.occupancy = occupancy
                    self.temp_factor = temp_factor
                    self.segment_id = segment_id
                    self.element = element
                    self.charge = charge

                    # Calculate the mass of the atom
                    self.mass = ATOMIC_WEIGHTS[element]
    
    def read_pdb(self, file_path, chains=None, name=None):
        """
        Read a PDB file and populate the ProteinStructure object with its contents. Atom elements are obtained from the
        atom name. 
        
        Note: The element is determined from columns 13-14 of the PDB file, so check that your PDB file is formatted
        correctly. Specifically, hydrogen atoms can sometimes have weird names such as 'HE11', and this function would
        incorrectly assign helium as the element. If you encounter this issue, you can manually set the element of the
        atom after reading the PDB file. 
        
        Reference: https://cdn.rcsb.org/wwpdb/docs/documentation/file-format/PDB_format_1992.pdf
        
        Parameters
        ----------
        file_path : str
            The path to the PDB file.
        chains : str or list of str, optional
            The chain(s) to read from the PDB file. If None, read all chains. Defaults to None.
        name : str, optional
            The name of the protein structure. Defaults to None.
        """
        self.name = name # The name of the protein structure. Defaults to None.

        # If the user passed a string for chains, convert it to a list
        if isinstance(chains, str):
            chains = [chains]

        # Parse PDB file and populate self.atoms
        p = Path(file_path)
        if p.suffix == ".pdb":
            with open(p, "r") as f:
                for line in f:
                    if line.startswith("ATOM"):

                        chain_name = line[21].strip()
                        # Check if the chain already exists in self.chains
                        chain = next((c for c in self.chains if c.name == chain_name), None)
                        if chains is None or chain_name in chains:
                            if chain is None:
                                chain = self.Chain(chain_name)
                                self.chains.append(chain)

                            # Check if the residue already exists in self.residues
                            res_name = line[17:20].strip()
                            res_seq = int(line[22:26].strip())
                            i_code = line[26].strip()
                            residue = next((r for r in chain.residues if r.res_seq == res_seq and r.i_code == i_code), None)
                            if residue is None:
                                # Create a Residue object and append it to self.residues
                                residue = self.Chain.Residue(res_name, res_seq, i_code)
                                # Check if the residue index is already assigned, if not assign it
                                if residue.index is None:
                                    # assign it next index value in the chain
                                    residue.index = len(chain.residues)
                                chain.residues.append(residue)

                            # Check if the atom already exists in self.atoms
                            atom_id = int(line[6:11].strip())
                            element = line[12:14].strip()
                            atom_name = line[12:16].strip()
                            alt_loc = line[16].strip()
                            x = float(line[30:38].strip())
                            y = float(line[38:46].strip())
                            z = float(line[46:54].strip())
                            occupancy = float(line[54:60].strip())
                            temp_factor = float(line[60:66].strip())
                            
                            segment_id = line[72:76].strip()
                            charge = line[78:80].strip()

                            # remove digits from element
                            element = ''.join([i for i in element if not i.isdigit()])

                            atom = next((a for a in residue.atoms if a.atom_id == atom_id), None)
                            if atom is None:
                                # Create an Atom object and append it to self.atoms
                                atom = self.Chain.Residue.Atom(atom_id, atom_name, alt_loc, x, y, z, occupancy, temp_factor, segment_id, element, charge)
                                residue.atoms.append(atom)

    
    def generate_coordinates(self):
        """
        Generate a dictionary of coordinates from the Structure object.

        Returns
        -------
        dict
            A dictionary of coordinates. The keys are the chain names and the values are the coordinate data.
            Each coordinate is a 4-element array [x, y, z, b], where x, y, and z are the coordinates, and b is a binary
            flag indicating whether the atom is a backbone atom (1) or not (0).
        """
        coordinates = {}
        for chain in self.structure:
            chain_coordinates = []
            for residue in chain:
                for atom in residue:
                    is_backbone = 1 if atom.name in ['N', 'CA', 'C', 'O'] else 0
                    chain_coordinates.append(np.append(atom.coord, is_backbone))
            coordinates[chain.id] = np.array(chain_coordinates)
        return coordinates 
    
    def get_sequence_dict(self):
        """
        Return the sequence of the protein structure as a dictionary of chains and sequences.

        :returns: A dictionary of chains and sequences. The keys are the chain names and the values are the sequences.
        :rtype: dict
        """
        sequences = {}
        for chain in self.chains:
            sequence = ""
            for residue in chain.residues:
                sequence += THREE_TO_ONE[residue.name]
            sequences[chain.name] = sequence
        return sequences
    
    def write_coordinates_hdf5(file_path, coordinates):
        """
        Write coordinate data to an HDF5 file.

        Parameters
        ----------
        file_path : str
            The path to the HDF5 file.
        coordinates : dict
            A dictionary of coordinates. The keys are the chain names and the values are the coordinate data.
        """
        with h5py.File(file_path, 'w') as f:
            for chain_name, chain_coordinates in coordinates.items():
                f.create_dataset(chain_name, data=chain_coordinates)

    def read_coordinates_hdf5(file_path):
        """
        Read coordinate data from an HDF5 file.

        Parameters
        ----------
        file_path : str
            The path to the HDF5 file.

        Returns
        -------
        dict
            A dictionary of coordinates. The keys are the chain names and the values are the coordinate data.
        """
        coordinates = {}
        with h5py.File(file_path, 'r') as f:
            for chain_name in f.keys():
                coordinates[chain_name] = f[chain_name][:]
        return coordinates



from setuptools import setup, find_packages

setup(
    name='protein-design-tools',
    version='0.1.28',
    author='Andrew Schaub',
    author_email='andrew.schaub@protonmail.com',
    description='A library of tools for protein design.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/drewschaub/protein-design-tools',
    license=' MIT License',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
    ],
    packages=['protein_design_tools'],
    package_dir={'protein_design_tools': 'protein_design_tools'},
    python_requires = ">=3.6",
    install_requires=['numpy', 'h5py'],  # List your project dependencies here
)
